This is a short demo for the Autumn 2023 CSE333 class to evaluate a pool of DNA oligos using Rust instead of C++ as we did in the previous lecture.  Note that this is a familiar task and that's why it was chosen.  However, see the official documentation on the main Rust website to see examples that make Rust standout in terms of robustness and safety compared with languages like C++, yet it remains very efficient and thus ideal for systems programming.

# Getting setup with Rust

Follow the installation instructions on the official [Rust website](https://www.rust-lang.org/learn/get-started) to get your toolchain setup.  A great place to start learning is ["The Book"](https://doc.rust-lang.org/book/) which is the official introductory book for Rust and available online for free.  Once you become more comfortable with basic Rust programming I recommend reading [Rust for Rustaceans](https://nostarch.com/rust-rustaceans).

# Building this crate

In the top-level directory run `cargo build`.  The first time this is run it may take a while as third-party crates will be automatically downloaded for you and compiled (ensure you have internet access).  By default, this will create an unoptimized 'debug' build.  To build an optimized version you want a 'release' build with `cargo build --release`.

This example library contains documentation that follows the official convention.  This allows us to build beautiful reference documentation easily.  Try running `cargo doc --open` to try it.

# Running our example program

Instead of creating a main executable file it is customary to provide example programs each demonstrating a small usecase for the crate in the `examples` directory.  We have created the equivalent of our C++ program in the `mydnapool` example program.  Run it with `cargo run --release --example mydnapool`.

# Extending this code

The following are worthwhile exercises as you continue your journey in learning Rust.

## Add command line processing to the example program

We hardcoded our poolsize in the example program.  Make the program accept arguments for number of oligos in the pool and oligo length.  The [clap](https://docs.rs/clap/latest/clap/) crate makes writing intuitive and well documented command line interfaces (CLIs) an easy task.

## Vectorized Hamming distance

Investigate how you can use [intrinsics](https://www.intel.com/content/www/us/en/docs/intrinsics-guide/index.html) [in Rust](https://doc.rust-lang.org/beta/core/arch/x86_64/index.html) to compute the Hamming distance between DNA oligos using vectorized instructions on the x86-64 platform, like we did earlier in C++.

Add a vectorized implementation using [AVX2](https://en.wikipedia.org/wiki/Advanced_Vector_Extensions) instructions.  Add functionality to detect, at runtime, if AVX2 is supported.  If not then delegate Hamming distance calculation to the non-vectorized code.


## Add benchmarks

Investigate how you can add benchmarks to evaluate the performance of different Hamming distance implementations.  [Criterion](https://bheisler.github.io/criterion.rs/book/getting_started.html) is a popular crate for microbenchmarks.  Once setup, you will be able to call `cargo bench`.


## Add additional functionality, document it, and test

Extend the simple DNA utility library with any new functionality you are interested in.  For example, generate random sequences within a specified [GC content](https://en.wikipedia.org/wiki/GC-content).  Be sure to add [unit tests](https://doc.rust-lang.org/rust-by-example/testing/unit_testing.html) integrated with the code you write.  Documentation and testing are not after thoughts; they are fully supported by the standard toolchain.  Note that our initial crate implementation made heavy use of [doctests](https://doc.rust-lang.org/rust-by-example/testing/doc_testing.html) that not only document and demonstrate how to use different parts of the crate with actual code examples, but are also tests that are run when `cargo test` is called.  This type of testing is excellent for ensuring the public interface of your crate (how users can iteract with it) is well documented and tested.

## Make the code idiomatic

Rust code can be formatted in a standard way with `cargo fmt`, and there is even a 'Clippy' tool to suggest how your code style can be improved.  (You are probably too young to remember the Clippy from Microsoft Word / Excel / etc :-)  Try running `cargo clippy` to check your code.

## Make your code parallel

Checking the Hamming distance between all pairs of oligos is [embarrassingly parallel](https://en.wikipedia.org/wiki/Embarrassingly_parallel).  Now that you are a concurrency expert, investigate how you can evaluate all pairwise Hamming distances faster using multiple threads.  You may want to investigate the [Crossbeam](https://docs.rs/crossbeam/latest/crossbeam/) crate in addition to the [standard library support for threading](https://doc.rust-lang.org/std/thread/).  Investigate the use of the [Rayon](concurrent code.) crate that makes it easy to convert sequential code into a parallel implementation.
