// examples/mydnapool.rs
//
// A simple example program using the dna library in order to generate and
// evaluate a set of DNA oligos.

use anyhow::Result; // For nicer handling of Result types.
use itertools::Itertools; // In order to use tuple_combinations.
use rayon::iter::ParallelIterator;
use rayon::prelude::ParallelBridge;

use std::time::{Duration, Instant};

fn main() -> Result<()> {
    use dna::align::hamming_distance;
    use dna::BaseSeq;
    use dna::DegBase;
    use dna::DegBaseSeq;

    // Create a specification for a degenerate DNA base sequence specification,
    // of length 64, where each position must contain an `N` base (i.e., any base).
    let spec = DegBaseSeq::new(&vec![DegBase::N; 64]);
    assert_eq!(spec.len(), 64);

    // Setup a pseudorandom number generator.
    let mut rng = rand::thread_rng();

    // Specify how many sequences we want in our pool.
    let poolsize = 2000;

    // Generate the pool of random sequences.
    let pool: Vec<BaseSeq> = (0..poolsize).map(|_| spec.rand_seq(&mut rng)).collect();
    assert_eq!(pool.len(), poolsize);

    // Calculate the minimum Hamming distance between any pair of sequences
    // in the pool.
    let start = Instant::now();
    let min_hd = pool
        .iter() // returns an iterator over elements in the Vec
        .tuple_combinations() // adapt to iterate over each pair of elements
        .flat_map(|(s1, s2)| hamming_distance(s1, s2)) // return the Hamming distance for each pair
        .min(); // find the minimum over each calculated Hamming distance
    let duration = start.elapsed();
    println!("Time elapsed for sequential code: {:?}", duration);

    let start = Instant::now();
    let min_hd = pool
        .iter() // returns an iterator over elements in the Vec
        .tuple_combinations() // adapt to iterate over each pair of elements
        .par_bridge()
        .flat_map(|(s1, s2)| hamming_distance(s1, s2)) // return the Hamming distance for each pair
        .min(); // find the minimum over each calculated Hamming distance
    let duration = start.elapsed();
    println!("Time elapsed for parallel code: {:?}", duration);

    if let Some(min_hd) = min_hd {
        println!("Minimum pairwise Hamming distance: {}", min_hd);
    } else {
        println!("Generated sequence pool does not have a minimum pairwise Hamming distance.");
    }

    // Everything went well.
    Ok(())
}
