//! This module is used to calculate the [`sequence alignment`] between pairs
//! of DNA base sequences.
//!
//! In this simplified utility library there is support for calculating the [`Hamming distance`]
//! between a pair of [`crate::BaseSeq`]s.
//!
//! [`sequence alignment`]: https://en.wikipedia.org/wiki/Sequence_alignment
//! [`Hamming distance`]: https://en.wikipedia.org/wiki/Hamming_distance

use crate::BaseSeq;
use crate::DNAError;

/// Returns the Hamming distance between `s1` and `s2`.
///
/// # Errors
///
/// Will return [`DNAError::UndefinedDistance`] if `s1` and `s2` have different
/// lengths.
///
/// # Examples
///
/// ```rust
/// # fn main() -> Result<(), dna::DNAError> {
/// use std::str::FromStr; // In order to parse DNA from strings.
/// use dna::BaseSeq;
/// use dna::DNAError;
/// use dna::align::hamming_distance;
///
/// let o1 = BaseSeq::from_str("ACGTTGCA")?;
/// let o2 = BaseSeq::from_str("AAGTTCCG")?;
/// let o3 = BaseSeq::from_str("AAGTTCCGTTTT")?;
///
/// assert_eq!(hamming_distance(&o1, &o2), Ok(3));
/// assert_eq!(hamming_distance(&o2, &o3), Err(DNAError::UndefinedDistance));
/// # Ok(()) }
/// ```
pub fn hamming_distance(s1: &BaseSeq, s2: &BaseSeq) -> Result<usize, DNAError> {
    if s1.len() != s2.len() {
        return Err(DNAError::UndefinedDistance);
    }

    // Iterate over both sequences, in lock-step, filter those pairs that are
    // not equal, then count them.
    let distance = s1.iter().zip(s2.iter()).filter(|(a, b)| a != b).count();

    Ok(distance)
}
