//! This crate provides a (basic) utility library for working with DNA sequences.
//!
//! ```rust
//! use dna::Base;
//!
//! let bA = Base::A;
//! let bT = Base::T;
//!
//! assert_ne!(bA, bT);
//! ```

#![deny(missing_docs)]

use std::cmp::Ordering;
use std::convert::TryFrom;
use std::iter::FromIterator;
use std::ops::Neg;

use itertools::Itertools;
use thiserror::Error;

/// Errors that can result when using this library.
#[derive(Error, Debug, PartialEq, Eq)]
pub enum DNAError {
    /// The character does not represent a valid DNA base.
    #[error("the char `{0}` could not be converted to a valid `Base`")]
    InvalidBase(char),
    /// The character does not represent a valid [IUPAC degenerate
    /// base](https://en.wikipedia.org/wiki/Nucleic_acid_notation#IUPAC_notation)
    /// symbol for DNA.
    #[error("the char `{0}` could not be converted to a valid `DegBase`")]
    InvalidDegBase(char),
    /// The intersection of two IUPAC degenerate base symbols do not specify a valid DNA base.
    #[error("the intersection of the IUPAC degenerate symbols `{0}` and `{1}` do not specify a valid DNA base.")]
    EmptyIntersection(char, char),
    /// The distance metric is undefined.  This occurs, for instance, when attempting to calculate
    /// the Hamming distance between two sequences of different lengths.
    #[error("the distance is undefined for the supplied arguments")]
    UndefinedDistance,
}

/// A `Base` is one of the four DNA nucleotides.
///
/// # Examples
///
/// ```rust
/// use dna::Base;
///
/// // An array of DNA `Base`s for the sequence "GATACA".
/// let x = [Base::G, Base::A, Base::T, Base::A, Base::C, Base::A];
///
/// // The same sequence with less typing.
/// use dna::Base::*;
/// let y = [G, A, T, A, C, A];
///
/// assert_eq!(x, y);
///
/// // `Base`s have a strict ordering.
/// assert!(A < C && C < G && G < T);
///
/// // A `Base` can be converted from a valid `char`.
/// use std::convert::TryFrom;
///
/// assert_eq!(Base::try_from('a'), Ok(A));
/// assert_eq!(Base::try_from('A'), Ok(A));
/// assert_eq!(Base::try_from('N'), Err(dna::DNAError::InvalidBase('N')));
/// ```
#[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash)]
#[repr(u8)]
pub enum Base {
    /// Adenine
    A = 0,
    /// Cystine
    C = 1,
    /// Guanine
    G = 2,
    /// Thymine
    T = 3,
}

impl Base {
    /// Returns a the [`Watson-Crick`] complement of this base.
    ///
    /// ```rust
    /// use dna::Base::*;
    ///
    /// assert_eq!(A.complement(), T);
    /// assert_eq!(T.complement(), A);
    /// assert_eq!(C.complement(), G);
    /// assert_eq!(G.complement(), C);
    ///
    /// let base = C;
    /// assert_eq!(base.complement(), G);
    ///
    ///
    /// // `-<Base>` is syntatic sugar for `<Base>.complement()`.
    /// assert_eq!(-T, A);
    /// assert_eq!(-G, C);
    /// assert_eq!(-C, G);
    /// assert_eq!(-A, T);
    /// ```
    ///
    /// [`Watson-Crick`]: https://en.wikipedia.org/wiki/Base_pair
    #[inline]
    pub fn complement(&self) -> Base {
        match self {
            Base::A => Base::T,
            Base::C => Base::G,
            Base::G => Base::C,
            Base::T => Base::A,
        }
    }
}

/// `Base` implements the `Neg` trait as syntatic sugar to mean
/// [`complement()`](enum.Base.html#method.complement_of).
impl Neg for Base {
    type Output = Self;

    #[inline]
    fn neg(self) -> Self::Output {
        self.complement()
    }
}

impl TryFrom<char> for Base {
    type Error = DNAError;

    /// Construct a [`Base`] from `value`.
    ///
    /// # Errors
    ///
    /// Will return [`DNAError::InvalidBase`] if `c` cannot be converted into a [`Base`].
    ///
    /// # Examples
    ///
    /// ```rust
    /// use dna::Base;
    /// use std::convert::TryFrom;
    ///
    /// assert_eq!(Base::try_from('a'), Ok(Base::A));
    /// assert_eq!(Base::try_from('A'), Ok(Base::A));
    /// assert_eq!(Base::try_from('N'), Err(dna::DNAError::InvalidBase('N')));
    /// ```
    #[inline]
    fn try_from(value: char) -> Result<Self, DNAError> {
        match value {
            'A' | 'a' => Ok(Base::A),
            'C' | 'c' => Ok(Base::C),
            'G' | 'g' => Ok(Base::G),
            'T' | 't' => Ok(Base::T),
            c => Err(DNAError::InvalidBase(c)),
        }
    }
}

impl From<Base> for char {
    #[inline]
    fn from(b: Base) -> char {
        match b {
            Base::A => 'A',
            Base::C => 'C',
            Base::G => 'G',
            Base::T => 'T',
        }
    }
}

impl std::fmt::Display for Base {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(f, "{}", char::from(*self))
    }
}

/// A `DegBase` is one of the [IUPAC degenerate base] symbols for DNA.
///
/// Degenerate base symbols are useful to specify sequence design constraints.
///
/// # Examples
///
/// ```rust
/// use dna::DegBase;
/// use dna::DegBase::*;
///
/// // A specification for an oligo sequence that begins with `TTT`, has a run
/// // of four *strong* bases, and ends with `AAA`.
/// let oligo_spec = [T, T, T, S, S, S, S, A, A, A];
///
/// // A `DegBase` can be converted from a valid `char`.
/// use std::convert::TryFrom;
///
/// assert_eq!(DegBase::try_from('a'), Ok(A));
/// assert_eq!(DegBase::try_from('S'), Ok(S));
/// assert_eq!(DegBase::try_from('z'), Err(dna::DNAError::InvalidDegBase('z')));
/// ```
///
/// [IUPAC degenerate base]: https://en.wikipedia.org/wiki/Nucleic_acid_notation#IUPAC_notation
#[derive(Clone, Copy, Debug, PartialEq, Eq, Hash)]
pub enum DegBase {
    /// Adenine
    A,
    /// Cystine
    C,
    /// Guanine
    G,
    /// Thymine
    T,
    /// Adenine or Guanine
    R,
    /// Cystine or Thymine
    Y,
    /// Cystine or Guanine
    S,
    /// Adenine or Thymine
    W,
    /// Guanine or Thymine
    K,
    /// Adenine or Cystine
    M,
    /// Cystine or Guanine or Thymine
    B,
    /// Adenine or Guanine or Thymine
    D,
    /// Adenine or Cystine or Thymine
    H,
    /// Adenine or Cystine or Guanine
    V,
    /// Adenine or Cystine or Guanine or Thymine
    N,
}

impl TryFrom<char> for DegBase {
    type Error = DNAError;

    fn try_from(value: char) -> Result<Self, DNAError> {
        match value {
            'A' | 'a' => Ok(DegBase::A),
            'C' | 'c' => Ok(DegBase::C),
            'G' | 'g' => Ok(DegBase::G),
            'T' | 't' => Ok(DegBase::T),
            'R' | 'r' => Ok(DegBase::R),
            'Y' | 'y' => Ok(DegBase::Y),
            'S' | 's' => Ok(DegBase::S),
            'W' | 'w' => Ok(DegBase::W),
            'K' | 'k' => Ok(DegBase::K),
            'M' | 'm' => Ok(DegBase::M),
            'B' | 'b' => Ok(DegBase::B),
            'D' | 'd' => Ok(DegBase::D),
            'H' | 'h' => Ok(DegBase::H),
            'V' | 'v' => Ok(DegBase::V),
            'N' | 'n' => Ok(DegBase::N),
            c => Err(DNAError::InvalidDegBase(c)),
        }
    }
}

impl From<DegBase> for char {
    fn from(b: DegBase) -> char {
        match b {
            DegBase::A => 'A',
            DegBase::C => 'C',
            DegBase::G => 'G',
            DegBase::T => 'T',
            DegBase::R => 'R',
            DegBase::Y => 'Y',
            DegBase::S => 'S',
            DegBase::W => 'W',
            DegBase::K => 'K',
            DegBase::M => 'M',
            DegBase::B => 'B',
            DegBase::D => 'D',
            DegBase::H => 'H',
            DegBase::V => 'V',
            DegBase::N => 'N',
        }
    }
}

impl std::fmt::Display for DegBase {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(f, "{}", char::from(*self))
    }
}

impl DegBase {
    /// Returns the degenerate base that denotes the set of DNA bases common to
    /// `b1` and `b2`.
    ///
    /// Returns [`None`] when the intersection is empty. See [IUPAC degenerate
    /// base] for details.
    ///
    /// # Examples
    /// ```rust
    /// use dna::DegBase;
    /// use dna::DegBase::*;
    ///
    /// assert_eq!(DegBase::intersection(M, Y), Some(C));
    /// assert_eq!(DegBase::intersection(N, A), Some(A));
    /// assert_eq!(DegBase::intersection(V, D), Some(R));
    /// assert_eq!(DegBase::intersection(W, S), None);
    /// ```
    /// [IUPAC degenerate base]: https://en.wikipedia.org/wiki/Nucleic_acid_notation#IUPAC_notation
    pub fn intersection(b1: DegBase, b2: DegBase) -> Option<DegBase> {
        use DegBase::*;

        // Handle equal case first to shorten match statement.
        if b1 == b2 {
            return Some(b1);
        }

        // The remainder assumes that we know the bases differ.
        match (b1, b2) {
            //  If N is one, the intersection is the other.
            (_, N) => Some(b1),
            (N, _) => Some(b2),

            // Intersections that produce A.
            (A, R) | (R, A) | (A, W) | (W, A) | (A, M) | (M, A) => Some(A),
            (A, D) | (D, A) | (A, H) | (H, A) | (A, V) | (V, A) => Some(A),
            (R, W) | (W, R) | (R, M) | (M, R) | (R, H) | (H, R) => Some(A),
            (W, M) | (M, W) | (W, V) | (V, W) => Some(A),
            (M, D) | (D, M) => Some(A),
            // Intersections that produce C.
            (C, Y) | (Y, C) | (C, S) | (S, C) | (C, M) | (M, C) => Some(C),
            (C, B) | (B, C) | (C, H) | (H, C) | (C, V) | (V, C) => Some(C),
            (Y, S) | (S, Y) | (Y, M) | (M, Y) | (Y, V) | (V, Y) => Some(C),
            (S, M) | (M, S) | (S, H) | (H, S) => Some(C),
            (M, B) | (B, M) => Some(C),
            // Intersections that produce G.
            (G, R) | (R, G) | (G, S) | (S, G) | (G, K) | (K, G) => Some(G),
            (G, B) | (B, G) | (G, D) | (D, G) | (G, V) | (V, G) => Some(G),
            (R, S) | (S, R) | (R, K) | (K, R) | (R, B) | (B, R) => Some(G),
            (S, K) | (K, S) | (S, D) | (D, S) => Some(G),
            (K, V) | (V, K) => Some(G),
            // Intersections that produce T.
            (T, Y) | (Y, T) | (T, W) | (W, T) | (T, K) | (K, T) => Some(T),
            (T, B) | (B, T) | (T, D) | (D, T) | (T, H) | (H, T) => Some(T),
            (Y, W) | (W, Y) | (Y, K) | (K, Y) | (Y, D) | (D, Y) => Some(T),
            (W, K) | (K, W) | (W, B) | (B, W) => Some(T),
            (K, H) | (H, K) => Some(T),
            // Intersections that produce R.
            (R, D) | (D, R) | (R, V) | (V, R) => Some(R),
            (D, V) | (V, D) => Some(R),
            // Intersections that produce Y.
            (Y, B) | (B, Y) | (Y, H) | (H, Y) => Some(Y),
            (B, H) | (H, B) => Some(Y),
            // Intersections that produce S.
            (S, B) | (B, S) | (S, V) | (V, S) => Some(S),
            (B, V) | (V, B) => Some(S),
            // Intersections that produce W.
            (W, D) | (D, W) | (W, H) | (H, W) => Some(W),
            (D, H) | (H, D) => Some(W),
            // Intersections that produce K.
            (K, B) | (B, K) | (K, D) | (D, K) => Some(K),
            (B, D) | (D, B) => Some(K),
            // Intersections that produce M.
            (M, H) | (H, M) | (M, V) | (V, M) => Some(M),
            (H, V) | (V, H) => Some(M),
            // Every other intersection should be empty.
            _ => None,
        }
    }

    /// Tries to replace `self` with a degenerate base common to `self` and
    /// `other` and return the new value.
    ///
    /// # Errors
    ///
    /// If the intersection is empty then an error is returned.
    ///
    /// # Examples
    ///
    /// ```rust
    /// use dna::DegBase;
    /// use dna::DegBase::*;
    ///
    /// let mut dbase = DegBase::M;
    /// dbase.try_intersect(Y);
    /// assert_eq!(dbase, C);
    ///
    /// assert_eq!(N.try_intersect(A), Ok(A));
    /// assert_eq!(V.try_intersect(D), Ok(R));
    /// assert_eq!(W.try_intersect(S), Err(dna::DNAError::EmptyIntersection('W', 'S')));
    /// ```
    pub fn try_intersect(&mut self, other: DegBase) -> Result<DegBase, DNAError> {
        *self = DegBase::intersection(*self, other)
            .ok_or_else(|| DNAError::EmptyIntersection((*self).into(), other.into()))?;
        Ok(*self)
    }

    /// Returns an array of DNA bases represented by this degenerate base.
    pub fn bases(self) -> &'static [Base] {
        match self {
            DegBase::A => &[Base::A],
            DegBase::C => &[Base::C],
            DegBase::G => &[Base::G],
            DegBase::T => &[Base::T],
            DegBase::R => &[Base::A, Base::G],
            DegBase::Y => &[Base::C, Base::T],
            DegBase::S => &[Base::C, Base::G],
            DegBase::W => &[Base::A, Base::T],
            DegBase::K => &[Base::G, Base::T],
            DegBase::M => &[Base::A, Base::C],
            DegBase::B => &[Base::C, Base::G, Base::T],
            DegBase::D => &[Base::A, Base::G, Base::T],
            DegBase::H => &[Base::A, Base::C, Base::T],
            DegBase::V => &[Base::A, Base::C, Base::G],
            DegBase::N => &[Base::A, Base::C, Base::G, Base::T],
        }
    }

    /// Determines if the set of bases represented by `self` includes `b`.
    ///
    /// # Examples
    ///
    /// ```rust
    /// use dna::Base;
    /// use dna::DegBase;
    ///
    /// let dbase = DegBase::S;
    /// assert_eq!(dbase.valid_base(Base::C), true);
    /// assert_eq!(dbase.valid_base(Base::G), true);
    /// assert_eq!(dbase.valid_base(Base::A), false);
    /// assert_eq!(dbase.valid_base(Base::T), false);
    /// ```
    pub fn valid_base(self, b: Base) -> bool {
        self.bases().iter().any(|x| *x == b)
    }

    /// Returns a random DNA base represented by this degenerate base.
    ///
    /// # Examples
    ///
    /// ```rust
    /// use dna::Base;
    /// use dna::DegBase;
    /// use rand;
    ///
    /// // Setup a pseudorandom number generator.
    /// let mut rng = rand::thread_rng();
    ///
    /// let dbase = DegBase::S;
    /// let base = dbase.rand_base(&mut rng);
    /// assert!( base == Base::C || base == Base::G );
    ///
    /// let dbase = DegBase::W;
    /// let base = dbase.rand_base(&mut rng);
    /// assert!( base == Base::A || base == Base::T );
    /// ```
    pub fn rand_base<R>(self, rng: &mut R) -> Base
    where
        R: rand::Rng,
    {
        use rand::seq::SliceRandom;
        *self.bases().choose(rng).unwrap()
    }
}

/// A sequence of DNA [`Base`]s.
#[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord)]
pub struct BaseSeq(smallvec::SmallVec<[Base; 64]>);

impl FromIterator<Base> for BaseSeq {
    fn from_iter<I: IntoIterator<Item = Base>>(iter: I) -> Self {
        BaseSeq(smallvec::SmallVec::from_iter(iter))
    }
}

impl std::str::FromStr for BaseSeq {
    type Err = <Base as TryFrom<char>>::Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        s.chars().map(Base::try_from).collect()
    }
}

impl std::fmt::Display for BaseSeq {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(f, "{}", self.0.iter().join(""))
    }
}

impl std::convert::AsRef<[Base]> for BaseSeq {
    fn as_ref(&self) -> &[Base] {
        &self.0
    }
}

impl IntoIterator for BaseSeq {
    type Item = Base;
    type IntoIter = smallvec::IntoIter<[Base; 64]>;

    fn into_iter(self) -> Self::IntoIter {
        self.0.into_iter()
    }
}

impl BaseSeq {
    /// Create a base sequence from an slice of [`Base`].
    ///
    /// # Examples
    ///
    /// ```rust
    /// use dna::Base::*;
    /// use dna::BaseSeq;
    ///
    /// let seq1 = BaseSeq::new(&[G, A, T, A, C, A]);
    /// // A `BaseSeq` is also a slice of `Base`.
    /// let seq2 = BaseSeq::new(&seq1);
    ///
    /// assert_eq!(seq1, seq2);
    /// ```
    pub fn new<B: AsRef<[Base]>>(seq: B) -> BaseSeq {
        let mut bs = BaseSeq::empty();
        bs.append(seq);
        bs
    }

    /// Create an empty base sequence.
    ///
    /// # Examples
    ///
    /// ```rust
    /// use dna::BaseSeq;
    ///
    /// let seq1 = BaseSeq::empty();
    /// let seq2 = BaseSeq::new(&[]);
    /// assert_eq!(seq1, seq2);
    /// ```
    pub fn empty() -> BaseSeq {
        BaseSeq(smallvec::smallvec![])
    }

    /// Returns the reverse [`Watson-Crick`] complement of this base sequence.
    ///
    /// ```rust
    /// use dna::Base::*;
    /// use dna::BaseSeq;
    ///
    /// let seq = BaseSeq::new(&[G, A, T, A, C, A]);
    /// let rc = BaseSeq::new(&[T, G, T, A, T, C]);
    ///
    /// assert_eq!(seq.reverse_complement(), rc);
    /// assert_eq!(rc.reverse_complement(), seq);
    /// ```
    /// [`Watson-Crick`]: https://en.wikipedia.org/wiki/Base_pair
    #[inline]
    pub fn reverse_complement(&self) -> BaseSeq {
        self.as_ref().iter().rev().map(Base::complement).collect()
    }

    /// The length of the base sequence.
    ///
    /// # Examples
    ///
    /// ```rust
    /// use dna::BaseSeq;
    /// use dna::Base::*;
    ///
    /// let mut seq = BaseSeq::new(&[T, G, T, A, T, C]);
    /// assert_eq!(seq.len(), 6);
    ///
    /// seq.append(&[T, T]);
    /// assert_eq!(seq.len(), 8);
    /// ```
    pub fn len(&self) -> usize {
        self.0.len()
    }

    /// Returns true if `self` has a sequence of length zero.
    ///
    /// # Examples
    ///
    /// ```rust
    /// use dna::BaseSeq;
    /// use dna::Base::*;
    ///
    /// let empty = BaseSeq::new(&[]);
    /// assert_eq!(empty.is_empty(), true);
    ///
    /// let non_empty = BaseSeq::new(&[G, A, T, A, C, A]);
    /// assert_eq!(non_empty.is_empty(), false);
    /// ```
    pub fn is_empty(&self) -> bool {
        self.len() == 0
    }

    /// Appends `other` to the 3' (right) end of the sequence.
    ///
    /// # Examples
    ///
    /// ```rust
    /// # fn main() -> Result<(), dna::DNAError> {
    /// use std::str::FromStr;
    /// use dna::BaseSeq;
    /// use dna::Base::*;
    ///
    /// let mut seq1 = BaseSeq::empty();
    /// let seq2 = BaseSeq::from_str("TGTATC")?;
    ///
    /// seq1.append(&[T, G, T, A, T, C]);
    /// assert_eq!(seq1, seq2);
    ///
    /// let seq3 = BaseSeq::from_str("TTT")?;
    /// seq1.append(&seq3);
    /// assert_ne!(seq1, seq2);
    /// assert_ne!(seq1, seq3);
    /// assert_eq!(seq1, BaseSeq::from_str("TGTATCTTT")?);
    /// # Ok(()) }
    /// ```
    pub fn append<B: AsRef<[Base]>>(&mut self, other: B) {
        self.0.extend_from_slice(other.as_ref());
    }

    /// Returns an iterator over each length `k` window of the [`Base`]
    /// sequence.
    ///
    /// # Examples
    ///
    /// ```rust
    /// use dna::Base::*;
    /// use dna::BaseSeq;
    ///
    /// let seq = BaseSeq::new(&[A, C, G, T, A, C, G, T]);
    /// let expected = vec![
    ///     &[A, C],
    ///     &[C, G],
    ///     &[G, T],
    ///     &[T, A],
    ///     &[A, C],
    ///     &[C, G],
    ///     &[G, T],
    /// ];
    /// for (idx, kmer) in seq.kmers(2).enumerate() {
    ///     assert_eq!(expected[idx], kmer);
    /// }
    /// ```
    pub fn kmers(&self, k: usize) -> KmerIter {
        KmerIter::from(self, k)
    }

    /// Returns an iterator over each [`Base`] in the sequence.
    ///
    /// # Examples
    ///
    /// ```rust
    /// use dna::Base::*;
    /// use dna::BaseSeq;
    ///
    /// let seq = BaseSeq::new(&[A, C, G, T, A, C, G, T]);
    ///
    /// let cnt_A = seq.iter().filter(|&base| base == A).count();
    /// assert_eq!(cnt_A, 2);
    ///
    /// let mut iter = seq.iter();
    /// assert_eq!(iter.next(), Some(A));
    /// ```
    pub fn iter(&self) -> BaseIter {
        self.0.as_ref().iter().cloned()
    }
}

type BaseIter<'a> = std::iter::Cloned<std::slice::Iter<'a, Base>>;

/// An iterator over each length `k` window of the [`Base`]
/// sequence.
///
/// This struct is created by the [`kmers`] method on [`BaseSeq`]. See its
/// documentation for more.
///
/// [`kmers`]: BaseSeq::kmers
#[derive(Debug)]
pub struct KmerIter<'a> {
    // bs: &'a BaseSeq,
    // k: usize,
    kmers: std::slice::Windows<'a, Base>,
}

impl<'a> KmerIter<'a> {
    fn from(bs: &'a BaseSeq, k: usize) -> KmerIter<'a> {
        let kmers = bs.as_ref().windows(k);
        // KmerIter { bs, k, kmers }
        KmerIter { kmers }
    }
}

impl<'a> Iterator for KmerIter<'a> {
    type Item = &'a [Base];

    fn next(&mut self) -> Option<Self::Item> {
        self.kmers.next()
    }
}

/// A sequence of [`DegBase`]s.
#[derive(Clone, Debug, PartialEq, Eq, Hash)]
pub struct DegBaseSeq(smallvec::SmallVec<[DegBase; 64]>);

impl std::iter::FromIterator<DegBase> for DegBaseSeq {
    fn from_iter<I: IntoIterator<Item = DegBase>>(iter: I) -> Self {
        DegBaseSeq(smallvec::SmallVec::from_iter(iter))
    }
}

impl std::str::FromStr for DegBaseSeq {
    type Err = <DegBase as TryFrom<char>>::Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        s.chars().map(DegBase::try_from).collect()
    }
}

impl std::fmt::Display for DegBaseSeq {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        write!(f, "{}", self.0.iter().join(""))
    }
}

impl std::convert::AsRef<[DegBase]> for DegBaseSeq {
    fn as_ref(&self) -> &[DegBase] {
        &self.0
    }
}

impl DegBaseSeq {
    /// Create a degenerate base sequence from an slice of [`DegBase`].
    ///
    /// # Examples
    ///
    /// ```rust
    /// use dna::DegBase::*;
    /// use dna::DegBaseSeq;
    ///
    /// let dseq1 = DegBaseSeq::new(&[G, A, N, A, C, A]);
    /// // A `DegBaseSeq` is also a slice of `DegBase`.
    /// let dseq2 = DegBaseSeq::new(&dseq1);
    ///
    /// assert_eq!(dseq1, dseq2);
    /// ```
    pub fn new<B: AsRef<[DegBase]>>(seq: B) -> DegBaseSeq {
        let mut bs = DegBaseSeq::empty();
        bs.append(seq);
        bs
    }

    /// Create an empty degenerate base sequence.
    ///
    /// # Examples
    ///
    /// ```rust
    /// use dna::DegBaseSeq;
    ///
    /// let dseq1 = DegBaseSeq::empty();
    /// let dseq2 = DegBaseSeq::new(&[]);
    /// assert_eq!(dseq1, dseq2);
    /// ```
    pub fn empty() -> DegBaseSeq {
        DegBaseSeq(smallvec::smallvec![])
    }

    /// The length of the degenerate base sequence.
    ///
    /// # Examples
    ///
    /// ```rust
    /// use dna::DegBaseSeq;
    /// use dna::DegBase::*;
    ///
    /// let mut dseq = DegBaseSeq::new(&[T, G, W, A, T, R]);
    /// assert_eq!(dseq.len(), 6);
    ///
    /// dseq.append(&[T, V]);
    /// assert_eq!(dseq.len(), 8);
    /// ```
    pub fn len(&self) -> usize {
        self.0.len()
    }

    /// Returns true if `self` has a sequence of length zero.
    ///
    /// # Examples
    ///
    /// ```rust
    /// use dna::DegBaseSeq;
    /// use dna::DegBase::*;
    ///
    /// let empty = DegBaseSeq::new(&[]);
    /// assert_eq!(empty.is_empty(), true);
    ///
    /// let non_empty = DegBaseSeq::new(&[N, A, Y, A, V, A]);
    /// assert_eq!(non_empty.is_empty(), false);
    /// ```
    pub fn is_empty(&self) -> bool {
        self.len() == 0
    }

    /// Appends `other` to the 3' (right) end of the sequence.
    ///
    /// # Examples
    ///
    /// ```rust
    /// # fn main() -> Result<(), dna::DNAError> {
    /// use std::str::FromStr;
    /// use dna::DegBaseSeq;
    /// use dna::DegBase::*;
    ///
    /// let mut dseq1 = DegBaseSeq::empty();
    /// let dseq2 = DegBaseSeq::from_str("NGNANC")?;
    ///
    /// dseq1.append(&[N, G, N, A, N, C]);
    /// assert_eq!(dseq1, dseq2);
    ///
    /// let dseq3 = DegBaseSeq::from_str("HHS")?;
    /// dseq1.append(&dseq3);
    /// assert_ne!(dseq1, dseq2);
    /// assert_ne!(dseq1, dseq3);
    /// assert_eq!(dseq1, DegBaseSeq::from_str("NGNANCHHS")?);
    /// # Ok(()) }
    /// ```
    pub fn append<B: AsRef<[DegBase]>>(&mut self, other: B) {
        self.0.extend_from_slice(other.as_ref());
    }

    /// Replace with a degenerate base sequence common to `self` and `other`.
    ///
    /// # Errors
    ///
    /// Will return [`DNAError::EmptyIntersection`] if there is a index where
    /// `self` and `other` do not represent any common bases. Note this is trivially
    /// true if `self` and `other` have different lenghts.
    ///
    /// # Examples
    ///
    /// ```rust
    /// # fn main() -> Result<(), dna::DNAError> {
    /// use std::str::FromStr;
    /// use dna::DegBaseSeq;
    /// use dna::DNAError;
    ///
    /// let ds1 = DegBaseSeq::from_str("ATA")?;
    /// let mut ds2 = DegBaseSeq::from_str("NNN")?;
    /// assert_eq!(ds2.try_intersect(&ds1), Ok(()));
    /// assert_eq!(ds2, ds1);
    ///
    /// let mut ds2 = DegBaseSeq::from_str("SSS")?;
    /// assert_eq!(ds2.try_intersect(&ds1), Err(DNAError::EmptyIntersection('S', 'A')));
    ///
    /// # Ok(()) }
    /// ```
    pub fn try_intersect(&mut self, other: &DegBaseSeq) -> Result<(), DNAError> {
        match self.len().cmp(&other.len()) {
            Ordering::Less => Err(DNAError::EmptyIntersection(
                ' ',
                char::from(other.as_ref()[self.len()]),
            )),
            Ordering::Greater => Err(DNAError::EmptyIntersection(
                char::from(other.as_ref()[other.len()]),
                ' ',
            )),
            Ordering::Equal => {
                for (b1, b2) in self.0.iter_mut().zip(other.0.iter()) {
                    b1.try_intersect(*b2).map_err(|_| {
                        DNAError::EmptyIntersection(char::from(*b1), char::from(*b2))
                    })?;
                }
                Ok(())
            }
        }
    }

    /// Returns true if `bs` is a DNA sequence specified by `self`.
    ///
    /// # Examples
    ///
    /// ```rust
    /// # fn main() -> Result<(), dna::DNAError> {
    /// use std::str::FromStr;
    /// use dna::BaseSeq;
    /// use dna::DegBaseSeq;
    ///
    /// let spec = DegBaseSeq::from_str("SSATW")?;
    /// let bs1 = BaseSeq::from_str("CGATA")?;
    /// let bs2 = BaseSeq::from_str("CGATAA")?;
    /// let bs3 = BaseSeq::from_str("TGATA")?;
    /// let bs4 = BaseSeq::from_str("GCATT")?;
    ///
    /// assert_eq!(spec.valid_seq(&bs1), true);
    /// assert_eq!(spec.valid_seq(&bs2), false);
    /// assert_eq!(spec.valid_seq(&bs3), false);
    /// assert_eq!(spec.valid_seq(&bs4), true);
    /// # Ok(()) }
    /// ```
    pub fn valid_seq(&self, bs: &BaseSeq) -> bool {
        if self.len() != bs.len() {
            false
        } else {
            self.0
                .iter()
                .zip(bs.as_ref().iter())
                .all(|(db, b)| db.valid_base(*b))
        }
    }

    /// Returns the number of DNA sequences belonging to the class specified by
    /// `self`.
    ///
    /// # Examples
    ///
    /// ```rust
    /// use dna::DegBaseSeq;
    /// use dna::DegBase::*;
    ///
    /// assert_eq!(DegBaseSeq::new(&[A, T, C, G]).seq_count(), 1);
    /// assert_eq!(DegBaseSeq::new(&[N, T, C, G]).seq_count(), 4);
    /// assert_eq!(DegBaseSeq::new(&[N, T, C, N]).seq_count(), 16);
    /// assert_eq!(DegBaseSeq::new(&[N, W, C, N]).seq_count(), 32);
    /// assert_eq!(DegBaseSeq::new(&[N, W, S, N]).seq_count(), 64);
    /// ```
    pub fn seq_count(&self) -> usize {
        self.0.iter().map(|&dg| dg.bases().len()).product()
    }

    /// Returns a random DNA base sequences represented by the class specified
    /// by `self`.
    ///
    /// # Examples
    ///
    /// ```rust
    /// use rand;
    /// use dna::DegBase::*;
    /// use dna::DegBaseSeq;
    ///
    /// // Setup a pseudorandom number generator.
    /// let mut rng = rand::thread_rng();
    ///
    /// let spec = DegBaseSeq::new(&[S, S, A, T, W]);
    /// let seq = spec.rand_seq(&mut rng);
    /// assert_eq!(spec.valid_seq(&seq), true);
    /// ```
    pub fn rand_seq<R>(&self, rng: &mut R) -> BaseSeq
    where
        R: rand::Rng,
    {
        self.0.iter().map(|db| db.rand_base(rng)).collect()
    }
}

// Organize our alignment routines into its own sub-module.
pub mod align;
